!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
MODULE MODI_ICE4_NUCLEATION
INTERFACE
SUBROUTINE ICE4_NUCLEATION(KSIZE, ODSOFT, ODCOMPUTE, &
                           PTHT, PPABST, PRHODREF, PEXN, PLSFACT, PT, &
                           PRVT, &
                           PCIT, PRVHENI_MR, PB_TH, PB_RV, PB_RI)
IMPLICIT NONE
INTEGER,                  INTENT(IN)    :: KSIZE
LOGICAL,                  INTENT(IN)    :: ODSOFT
LOGICAL, DIMENSION(KSIZE),INTENT(IN)    :: ODCOMPUTE
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PTHT    ! Theta at t
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PEXN    ! Exner function
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PLSFACT
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PT      ! Temperature at time t
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PCIT    ! Pristine ice n.c. at t
REAL, DIMENSION(KSIZE),   INTENT(OUT)   :: PRVHENI_MR ! Mixing ratio change due to the heterogeneous nucleation
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PB_TH
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PB_RV
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PB_RI
END SUBROUTINE ICE4_NUCLEATION
END INTERFACE
END MODULE MODI_ICE4_NUCLEATION
SUBROUTINE ICE4_NUCLEATION(KSIZE, ODSOFT, ODCOMPUTE, &
                           PTHT, PPABST, PRHODREF, PEXN, PLSFACT, PT, &
                           PRVT, &
                           PCIT, PRVHENI_MR, PB_TH, PB_RV, PB_RI)
!!
!!**  PURPOSE
!!    -------
!!      Computes the nucleation
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XALPI,XALPW,XBETAI,XBETAW,XGAMI,XGAMW,XMD,XMV,XTT
USE MODD_PARAM_ICE,      ONLY: LFEEDBACKT
USE MODD_RAIN_ICE_PARAM, ONLY: XALPHA1,XALPHA2,XBETA1,XBETA2,XMNU0,XNU10,XNU20
USE MODD_RAIN_ICE_DESCR, ONLY: XRTMIN
!
USE MODE_MPPDB
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                  INTENT(IN)    :: KSIZE
LOGICAL,                  INTENT(IN)    :: ODSOFT
LOGICAL, DIMENSION(KSIZE),INTENT(IN)    :: ODCOMPUTE
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PTHT    ! Theta at t
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PEXN    ! Exner function
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PLSFACT
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PT      ! Temperature at time t
REAL, DIMENSION(KSIZE),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PCIT    ! Pristine ice n.c. at t
REAL, DIMENSION(KSIZE),   INTENT(OUT)   :: PRVHENI_MR ! Mixing ratio change due to the heterogeneous nucleation
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PB_TH
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PB_RV
REAL, DIMENSION(KSIZE),   INTENT(INOUT) :: PB_RI
!
!$acc declare present(KSIZE,ODSOFT,ODCOMPUTE,PTHT,PPABST,PRHODREF,PEXN,PLSFACT,PT,PRVT,PCIT,PRVHENI_MR,PB_TH,PB_RV,PB_RI)
!
!*       0.2  declaration of local variables
!
INTEGER                   :: JI
LOGICAL                   :: GDSOFT, GFEEDBACKT !Workaround of PGI bug with OpenACC (at least up to 18.10 version)
LOGICAL, DIMENSION(KSIZE) :: GNEGT  ! Test where to compute the HEN process
LOGICAL, DIMENSION(KSIZE) :: GWORK, GWORK2
REAL, DIMENSION(KSIZE)    :: ZW ! work array
REAL, DIMENSION(KSIZE)    :: ZZW,      & ! Work array
                             ZUSW,     & ! Undersaturation over water
                             ZSSI        ! Supersaturation over ice
!
!$acc declare create(JI,GNEGT,GWORK,GWORK2,ZW,ZZW,ZUSW,ZSSI)
!
!$acc declare copyin(XALPI,XALPW,XBETAI,XBETAW,XGAMI,XGAMW,XMD,XMV,XTT, &
!$acc&               LFEEDBACKT,                                        &
!$acc&               XALPHA1,XALPHA2,XBETA1,XBETA2,XMNU0,XNU10,XNU20,   &
!$acc&               XRTMIN)
!
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PTHT,"ICE4_NUCLEATION beg:PTHT")
  CALL MPPDB_CHECK(PPABST,"ICE4_NUCLEATION beg:PPABST")
  CALL MPPDB_CHECK(PRHODREF,"ICE4_NUCLEATION beg:PRHODREF")
  CALL MPPDB_CHECK(PEXN,"ICE4_NUCLEATION beg:PEXN")
  CALL MPPDB_CHECK(PLSFACT,"ICE4_NUCLEATION beg:PLSFACT")
  CALL MPPDB_CHECK(PT,"ICE4_NUCLEATION beg:PT")
  CALL MPPDB_CHECK(PRVT,"ICE4_NUCLEATION beg:PRVT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"ICE4_NUCLEATION beg:PCIT")
  CALL MPPDB_CHECK(PB_TH,"ICE4_NUCLEATION beg:PB_TH")
  CALL MPPDB_CHECK(PB_RV,"ICE4_NUCLEATION beg:PB_RV")
  CALL MPPDB_CHECK(PB_RI,"ICE4_NUCLEATION beg:PB_RI")
END IF
!
!Workaround of PGI bug with OpenACC (at least up to 18.10 version)
GFEEDBACKT = LFEEDBACKT
! !$acc kernels
!PW: bug?: for the moment I take the value from the CPU as OK. GPU value seems to be lost when this subroutine is called
GDSOFT = ODSOFT
! !$acc end kernels
!
!$acc kernels
PRVHENI_MR(:)=0.
IF(.NOT. GDSOFT) THEN
  GNEGT(:)=PT(:)<XTT .AND. PRVT>XRTMIN(1) .AND. ODCOMPUTE(:)
  PRVHENI_MR(:)=0.
  ZSSI(:)=0.
  ZUSW(:)=0.
  ZZW(:)=0.
  WHERE(GNEGT(:))
#ifndef MNH_BITREP
    ZZW(:)=ALOG(PT(:))
    ZUSW(:)=EXP(XALPW - XBETAW/PT(:) - XGAMW*ZZW(:))          ! es_w
    ZZW(:)=EXP(XALPI - XBETAI/PT(:) - XGAMI*ZZW(:))           ! es_i
#else
    ZZW(:)  = BR_LOG(PT(:))
    ZUSW(:) = BR_EXP(XALPW - XBETAW/PT(:) - XGAMW*ZZW(:))          ! es_w
    ZZW(:)  = BR_EXP(XALPI - XBETAI/PT(:) - XGAMI*ZZW(:))          ! es_i
#endif
  END WHERE
  WHERE(GNEGT(:))
    ZZW(:)=MIN(PPABST(:)/2., ZZW(:))             ! safety limitation
    ZSSI(:)=PRVT(:)*(PPABST(:)-ZZW(:)) / ((XMV/XMD)*ZZW(:)) - 1.0
                                                 ! Supersaturation over ice
    ZUSW(:)=MIN(PPABST(:)/2., ZUSW(:))            ! safety limitation
    ZUSW(:)=(ZUSW(:)/ZZW(:))*((PPABST(:)-ZZW(:))/(PPABST(:)-ZUSW(:))) - 1.0
                               ! Supersaturation of saturated water vapor over ice
    !
    !*       3.1     compute the heterogeneous nucleation source RVHENI
    !
    !*       3.1.1   compute the cloud ice concentration
    !
    ZSSI(:)=MIN(ZSSI(:), ZUSW(:)) ! limitation of SSi according to SSw=0
  END WHERE
  ZZW(:)=0.
  GWORK(:)  = GNEGT(:) .AND. PT(:)<XTT-5.0 .AND. ZSSI(:)>0.0
  GWORK2(:) = GNEGT(:) .AND. PT(:)<=XTT-2.0 .AND. PT(:)>=XTT-5.0 .AND. ZSSI(:)>0.0
#ifndef MNH_BITREP
#ifndef _OPENACC
  WHERE(GWORK(:))
    ZZW(:)=XNU20*EXP(XALPHA2*ZSSI(:)-XBETA2)
  ELSEWHERE(GWORK2(:))
    ZZW(:)=MAX(XNU20*EXP(-XBETA2 ), &
               XNU10*EXP(-XBETA1*(PT(:)-XTT))*(ZSSI(:)/ZUSW(:))**XALPHA1)
  END WHERE
#else
!$acc loop independent
  DO JI=1,KSIZE
    IF (GWORK(JI)) THEN
      ZZW(JI)=XNU20*EXP(XALPHA2*ZSSI(JI)-XBETA2)
    ELSE IF (GWORK2(JI)) THEN
      ZZW(JI)=MAX(XNU20*(-XBETA2 ), &
              XNU10*EXP(-XBETA1*(PT(JI)-XTT))*(ZSSI(JI)/ZUSW(JI)**XALPHA1)
    END IF
  END DO
#endif
#else
#ifndef _OPENACC
  WHERE(GWORK(:))
    ZZW(:)=XNU20*BR_EXP(XALPHA2*ZSSI(:)-XBETA2)
  ELSEWHERE(GWORK2(:))
    ZZW(:)=MAX(XNU20*BR_EXP(-XBETA2 ), &
               XNU10*BR_EXP(-XBETA1*(PT(:)-XTT))*BR_POW(ZSSI(:)/ZUSW(:),XALPHA1))
  END WHERE
#else
!$acc loop independent
  DO JI=1,KSIZE
    IF (GWORK(JI)) THEN
      ZZW(JI)=XNU20*BR_EXP(XALPHA2*ZSSI(JI)-XBETA2)
    ELSE IF (GWORK2(JI)) THEN
      ZZW(JI)=MAX(XNU20*BR_EXP(-XBETA2 ), &
              XNU10*BR_EXP(-XBETA1*(PT(JI)-XTT))*BR_POW(ZSSI(JI)/ZUSW(JI),XALPHA1))
    END IF
  END DO
#endif
#endif
  WHERE(GNEGT(:))
    ZZW(:)=ZZW(:)-PCIT(:)
    ZZW(:)=MIN(ZZW(:), 50.E3) ! limitation provisoire a 50 l^-1
  END WHERE
  WHERE(GNEGT(:))
    !
    !*       3.1.2   update the r_i and r_v mixing ratios
    !
    PRVHENI_MR(:)=MAX(ZZW(:), 0.0)*XMNU0/PRHODREF(:)
    PRVHENI_MR(:)=MIN(PRVT(:), PRVHENI_MR(:))
  END WHERE
  !Limitation due to 0 crossing of temperature
  IF(GFEEDBACKT) THEN
    ZW(:)=0.
    WHERE(GNEGT(:))
      ZW(:)=MIN(PRVHENI_MR(:), &
                MAX(0., (XTT/PEXN(:)-PTHT(:))/PLSFACT(:))) / &
                MAX(PRVHENI_MR(:), 1.E-20)
    END WHERE
  ELSE
    ZW(:)=1.
  ENDIF
  PRVHENI_MR(:)=PRVHENI_MR(:)*ZW(:)
  PCIT(:)=MAX(ZZW(:)*ZW(:)+PCIT(:), PCIT(:))
  !
  PB_RI(:)=PB_RI(:) + PRVHENI_MR(:)
  PB_RV(:)=PB_RV(:) - PRVHENI_MR(:)
  PB_TH(:)=PB_TH(:) + PRVHENI_MR(:)*PLSFACT(:)
ENDIF
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRVHENI_MR,"ICE4_NUCLEATION end:PRVHENI_MR")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"ICE4_NUCLEATION end:PCIT")
  CALL MPPDB_CHECK(PB_TH,"ICE4_NUCLEATION end:PB_TH")
  CALL MPPDB_CHECK(PB_RV,"ICE4_NUCLEATION end:PB_RV")
  CALL MPPDB_CHECK(PB_RI,"ICE4_NUCLEATION end:PB_RI")
END IF
!
END SUBROUTINE ICE4_NUCLEATION
